/* $Id: vircam_sky.c,v 1.10 2013-10-15 16:51:47 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2008 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:51:47 $
 * $Revision: 1.10 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cpl.h>
#include <math.h>
#include <string.h>

#include "vircam_sky.h"
#include "vircam_mask.h"
#include "vircam_fits.h"
#include "vircam_mods.h"
#include "vircam_utils.h"
#include "vircam_stats.h"
#include "vircam_wcsutils.h"
#include "vircam_pfits.h"

typedef struct {
    vir_fits **in;
    int      n;
    int      nalloc;
    int      status;
    vir_fits *xsky;
} xsky_struct;

static void vircam_sky_mask_grow(cpl_image *dith, float rad);
static float vircam_sky_med(vir_fits *sky);
static void vircam_sky_joffs(vir_fits **inlist, int nfiles, float **xoffs,
			     float **yoffs);
static void vircam_sky_xytoxy(cpl_wcs *inwcs, cpl_wcs *outwcs, 
			      cpl_matrix *inxy, cpl_matrix **outxy);

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Work out a masked sky estimate from an input jitter series

    \par Name:
        vircam_pawsky_mask
    \par Purpose:
        Work a masked sky estimate from an input jitter series
    \par Description:
        Given an input jitter series stars are iteratively masked and clipped
	to form a final sky estimate.
    \par Language:
        C
    \param inlist
        The input jitter series
    \param nfiles
        The number of images in the jitter series
    \param conf
        The input confidence map for the jitter series
    \param mask
        The bad pixel mask derived from the confidence map.
    \param skyout
        The output sky map
    \param niter
        The number of masking and clipping iterations to do
    \param ipix
        The minimum object size for masking
    \param thresh
        The threshold in sigma above sky for the object detection alrorithm
    \param nbsize
        The cell size for the background following part of the object 
	detection algorithm
    \param smkern
        The size in pixels of the smoothing kernal for the object detection
	algorithm
    \param status
        An input/output status that is the same as the returned values below.
    \retval VIR_OK
        if everything is ok
    \retval VIR_FATAL
        if something went wrong. An error message will appear.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_pawsky_mask(vir_fits **inlist, int nfiles, vir_fits *conf,
			      vir_mask *mask, vir_fits **skyout, int niter, 
			      int ipix, float thresh, int nbsize, float smkern, 
			      int *status) {
    int i,nx,ny,nbad0,lastone,iter,*confdata,npts,nbad_init,nbad,dbad;
    int xx1,xx2,yy1,yy2;
    const char *fctid = "vircam_pawsky_mask";
    cpl_image *dith,*dithc,*skyim,*im,*newim,*outim;
    cpl_mask *cplmask,*dithmask,*newmask,*curmask;
    vir_fits **list_ss,*dithf,*dithcf;
    unsigned char *inbpm,*rejmask,*rejplus;
    float medsky,*xoffs,*yoffs,mindx,mindy,fbad;
    cpl_propertylist *p,*drs;

    /* Inherited status */

    *skyout = NULL;
    if (*status != VIR_OK) 
	return(*status);

    /* If there aren't any images, then get out of here */

    if (nfiles == 0) {
	cpl_msg_error(fctid,"Sky correction impossible. No science frames");
	return(VIR_FATAL);
    }

    /* Wrap the input mask into a cpl_mask structure and use it to set the
       internal mask for each of the input images */

    inbpm = vircam_mask_get_data(mask);
    nx = vircam_mask_get_size_x(mask);
    ny = vircam_mask_get_size_y(mask);
    cplmask = cpl_mask_wrap((cpl_size)nx,(cpl_size)ny,(cpl_binary *)inbpm);
    for (i = 0; i < nfiles; i++)
	cpl_image_reject_from_mask(vircam_fits_get_image(inlist[i]),cplmask);
    cpl_mask_unwrap(cplmask);

    /* Do an initial sky combination */

    vircam_imcombine(inlist,nfiles,1,1,0,2.0,&outim,&rejmask,
		     &rejplus,&drs,status);
    *skyout = vircam_fits_wrap(outim,inlist[0],NULL,NULL);    
    freespace(rejmask);
    freespace(rejplus);
    freepropertylist(drs);

    /* Clean up any blank bits */

    (void)vircam_inpaint(*skyout,nbsize,status);

    /* If this is the only iteration then get out of here now */

    if (niter == 0)
	return(*status);

    /* Work out the median sky, ignoring the bad bits */

    medsky = vircam_sky_med(*skyout);

    /* Get jitter offsets from wcs */

    vircam_sky_joffs(inlist,nfiles,&xoffs,&yoffs);
    mindx = xoffs[0];
    mindy = yoffs[0];
    for (i = 1; i < nfiles; i++) {
	mindx = min(mindx,xoffs[i]);
	mindy = min(mindy,yoffs[i]);
    }

    /* Do an initial subtraction */

    list_ss = cpl_malloc(nfiles*sizeof(vir_fits *));
    skyim = vircam_fits_get_image(*skyout);
    for (i = 0; i < nfiles; i++) {
	im = vircam_fits_get_image(inlist[i]);
	newim = cpl_image_subtract_create(im,skyim);
	cpl_image_add_scalar(newim,(double)medsky);
	list_ss[i] = vircam_fits_wrap(newim,inlist[i],NULL,NULL);
    }
    vircam_fits_delete(*skyout);

    /* Set up some counters and begin the iteration loop */

    nbad0 = 0;
    lastone = 0;
    for (iter = 1; iter <= niter; iter++) {
	if (lastone)
	    break;
	lastone = (iter == niter);

	/* Dither the sky subtracted input images */

	(void)vircam_imdither(list_ss,&conf,nfiles,1,5.0,5.0,&p,&dith,
			      &dithc,status);
	cpl_propertylist_delete(p);

	/* How many dead pixels are there in the dithered image? */

	confdata = cpl_image_get_data_int(dithc);
	npts = (int)cpl_image_get_size_x(dithc)*(int)cpl_image_get_size_y(dithc);
	nbad_init = 0;
	for (i = 0; i < npts; i++)
	    if (confdata[i] == 0)
		nbad_init++;

	/* Get rid of the sky subtracted images */

	for (i = 0; i < nfiles; i++)
	    vircam_fits_delete(list_ss[i]);

	/* Wrap the result */

	dithf = vircam_fits_wrap(dith,inlist[0],NULL,NULL);
	dithcf = vircam_fits_wrap(dithc,conf,NULL,NULL);

	/* Now get an object mask from this dithered image */

	(void)vircam_opm(dithf,dithcf,ipix,thresh,nbsize,smkern,niter,status);

	/* How many flagged pixels are there, not counting the ones already
	   flagged in the confidence map */

	nbad = (int)cpl_image_count_rejected((const cpl_image *)dith) - 
	    nbad_init;
	dbad = nbad - nbad0;
	fbad = (iter > 1 ? (float)abs(dbad)/(float)nbad0 : 10000.0);
	cpl_msg_info(fctid,
		     "Iteration: %" CPL_SIZE_FORMAT ", Nreject: %" CPL_SIZE_FORMAT " %" CPL_SIZE_FORMAT,
		     (cpl_size)iter,(cpl_size)nbad,(cpl_size)nbad0);
	if (fbad < 0.025 || dbad < 0) {
	    lastone = 1;
	} else {
	    nbad0 = nbad;
	}

	/* Allow the mask to grow before the next iteration */

	vircam_sky_mask_grow(dith,2);

	/* Right, now decide which part of the mask belongs to each of the 
	   input images and update the input mask. */

	dithmask = cpl_image_get_bpm(dith);
	for (i = 0; i < nfiles; i++) {
	    xx1 = (int)(-mindx + xoffs[i] + 1.5);
	    xx2 = xx1 + nx - 1;
	    yy1 = (int)(-mindy + yoffs[i] + 1.5);
	    yy2 = yy1 + ny - 1;
	    newmask = cpl_mask_extract(dithmask,(cpl_size)xx1,(cpl_size)yy1,
				       (cpl_size)xx2,(cpl_size)yy2);
	    curmask = cpl_image_get_bpm(vircam_fits_get_image(inlist[i]));
	    cpl_mask_or(curmask,newmask);
	    cpl_mask_delete(newmask);
	}

	/* Delete the dither and its confidence map */
	
	vircam_fits_delete(dithf);
	vircam_fits_delete(dithcf);

	/* Do a new sky combination using these new object masks */

	vircam_imcombine(inlist,nfiles,1,1,0,2.0,&outim,&rejmask,
			 &rejplus,&drs,status);
        *skyout = vircam_fits_wrap(outim,inlist[0],NULL,NULL);    
	freespace(rejmask);
	freespace(rejplus);
	freepropertylist(drs);

        /* Clean up any blank bits */

        (void)vircam_inpaint(*skyout,nbsize,status);

	/* Do the sky subtraction again so long as this isn't 
	   the last iteration */

	if (! lastone) {
	    skyim = vircam_fits_get_image(*skyout);
	    for (i = 0; i < nfiles; i++) {
		im = vircam_fits_get_image(inlist[i]);
		newim = cpl_image_subtract_create(im,skyim);
		cpl_image_add_scalar(newim,(double)medsky);
		list_ss[i] = vircam_fits_wrap(newim,inlist[i],NULL,NULL);
	    }
	    vircam_fits_delete(*skyout);
	}
    }

    /* Free up some workspace that we've collected along the way */

    freespace(xoffs);
    freespace(yoffs);
    freespace(list_ss);
    return(*status);
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Work out a masked sky estimate from an input jitter series

    \par Name:
        vircam_pawsky_mask
    \par Purpose:
        Work a masked sky estimate from an input jitter series
    \par Description:
        Given an input jitter series stars are iteratively masked and clipped
	to form a final sky estimate.
    \par Language:
        C
    \param inlist
        The input jitter series
    \param nfiles
        The number of images in the jitter series
    \param conf
        The input confidence map for the jitter series
    \param mask
        The bad pixel mask derived from the confidence map.
    \param skyout
        The output sky map
    \param niter
        The number of masking and clipping iterations to do
    \param ipix
        The minimum object size for masking
    \param thresh
        The threshold in sigma above sky for the object detection alrorithm
    \param nbsize
        The cell size for the background following part of the object 
	detection algorithm
    \param smkern
        The size in pixels of the smoothing kernal for the object detection
	algorithm
    \param status
        An input/output status that is the same as the returned values below.
    \retval VIR_OK
        if everything is ok
    \retval VIR_FATAL
        if something went wrong. An error message will appear.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_pawsky_mask_pre(vir_fits **inlist, int nfiles, vir_fits *conf,
				  vir_mask *mask, vir_fits *objmask, 
				  int nbsize, vir_fits **skyout, int *status) {
    const char *fctid = "vircam_pawsky_mask_pre";
    unsigned char *inbpm,*rejplus,*rejmask;
    int nx,ny,npts,ncoor,i,j,nx_objmask,ny_objmask,k,kk,ix,iy,ind,*opm;
    cpl_mask *cplmask;
    cpl_wcs *wcs_objmask,*wcs;
    cpl_matrix *in_xy,*out_xy,*ddx,*ddx_out;
    cpl_image *im,*outim;
    cpl_propertylist *drs;
    double *inxy_data,xx,yy,*ddx_data,shiftx,shifty;

    /* Inherited status */

    *skyout = NULL;
    if (*status != VIR_OK) 
	return(*status);

    /* If there aren't any images, then get out of here */

    if (nfiles == 0) {
	cpl_msg_error(fctid,"Sky correction impossible. No science frames");
	return(VIR_FATAL);
    }

    /* Wrap the input mask into a cpl_mask structure and use it to set the
       internal mask for each of the input images */

    inbpm = vircam_mask_get_data(mask);
    nx = vircam_mask_get_size_x(mask);
    ny = vircam_mask_get_size_y(mask);
    cplmask = cpl_mask_wrap((cpl_size)nx,(cpl_size)ny,(cpl_binary *)inbpm);
    for (i = 0; i < nfiles; i++)
	cpl_image_reject_from_mask(vircam_fits_get_image(inlist[i]),cplmask);
    cpl_mask_unwrap(cplmask);

    /* Open the WCS of the object mask and get the pixel mask info */
    
    wcs_objmask = cpl_wcs_new_from_propertylist(vircam_fits_get_ehu(objmask));
    opm = cpl_image_get_data(vircam_fits_get_image(objmask));
    nx_objmask = cpl_image_get_size_x(vircam_fits_get_image(objmask));
    ny_objmask = cpl_image_get_size_y(vircam_fits_get_image(objmask));

    /* Get some memory so that we can do a transformation between the coordinate
       systems of the input frames and that of the object mask */
    
    npts = nx*ny;
    ncoor = 2;
    in_xy = cpl_matrix_new((cpl_size)npts,(cpl_size)ncoor);

    /* Initialise the data for the first x,y matrix */

    inxy_data = cpl_matrix_get_data(in_xy);
    k = 0;
    for (j = 0; j < ny; j++) {
	for (i = 0; i < nx; i++) {
	    inxy_data[k++] = (double)(i+1);
	    inxy_data[k++] = (double)(j+1);
	}
    }

    /* Initialise data for an offset matrix */

    ddx = cpl_matrix_new(1,2);
    ddx_data = cpl_matrix_get_data(ddx);
    ddx_data[0] = 1.0;
    ddx_data[1] = 1.0;

    /* Loop for each input file. Convert the xy matrix to an xy in the object
       mask frame of reference for the first file. Then use the offsets between
       the first image and the subsequent ones to define the location of
       the object pixels in the following images */

    for (k = 0; k < nfiles; k++) {
	wcs = cpl_wcs_new_from_propertylist(vircam_fits_get_ehu(inlist[k]));
	im = vircam_fits_get_image(inlist[k]);
	if (k == 0) 
	    vircam_sky_xytoxy(wcs,wcs_objmask,in_xy,&out_xy);
	vircam_sky_xytoxy(wcs,wcs_objmask,ddx,&ddx_out);
	cpl_wcs_delete(wcs);
	shiftx =  cpl_matrix_get(out_xy,0,0) - cpl_matrix_get(ddx_out,0,0);
	shifty =  cpl_matrix_get(out_xy,0,1) - cpl_matrix_get(ddx_out,0,1);
	kk = 0;
	for (j = 0; j < ny; j++) {
	    for (i = 0; i < nx; i++) {
		xx = cpl_matrix_get(out_xy,kk,0) - shiftx;
		yy = cpl_matrix_get(out_xy,kk,1) - shifty;
		kk++;
		ix = vircam_nint(xx);
		iy = vircam_nint(yy);
		if (ix < 1 || ix > nx_objmask  || iy < 1 || iy > ny_objmask)
		    continue;
		ind = (iy-1)*nx_objmask + ix - 1;
		if (opm[ind])
		    cpl_image_reject(im,i+1,j+1);
	    }
	}
	cpl_matrix_delete(ddx_out);
    }
    cpl_matrix_delete(in_xy);
    cpl_matrix_delete(out_xy);
    cpl_wcs_delete(wcs_objmask);

    /* Do an initial sky combination */

    vircam_imcombine(inlist,nfiles,1,1,0,2.0,&outim,&rejmask,
		     &rejplus,&drs,status);
    *skyout = vircam_fits_wrap(outim,inlist[0],NULL,NULL);    
    freespace(rejmask);
    freespace(rejplus);
    freepropertylist(drs);

    /* Clean up any blank bits */

    (void)vircam_inpaint(*skyout,nbsize,status);

    /* Get out of here */

    return(*status);
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Work sky estimate from an input tile series

    \par Name:
        vircam_tilesky
    \par Purpose:
        Work a sky estimate from an input tile
    \par Description:
        A list of images from a whole tile is given. These are broken up
	into groups that all have the same position within the jitter
	sequence. An intermediate sky is formed from each group and then a
	final output sky is formed from the intermediate sky files.
    \par Language:
        C
    \param inlist
        The input images from the tile
    \param nfiles
        The number of images in the jitter series
    \param mask
        The bad pixel mask derived from the confidence map.
    \param skyout
        The output sky map
    \param status
        An input/output status that is the same as the returned values below.
    \retval VIR_OK
        if everything is ok
    \retval VIR_FATAL
        if something went wrong. An error message will appear.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_tilesky(vir_fits **inlist, int nfiles, vir_mask *mask, 
			  vir_fits **skyout, int *status) { 
    int i,njit,nx,ny,ji,ns,npts;
    unsigned char *inbpm,*rejmask,*rejplus;
    cpl_mask *cplmask;
    xsky_struct *xskys;
    cpl_propertylist *drs;
    cpl_image *outim;
    vir_fits **xsky_fits;
    const char *fctid = "vircam_tilesky";

    /* Inherited status */

    *skyout = NULL;
    if (*status != VIR_OK) 
	return(*status);

    /* If there aren't any images, then get out of here */

    if (nfiles == 0) {
	cpl_msg_error(fctid,"Sky correction impossible. No science frames");
	return(VIR_FATAL);
    }

    /* Wrap the input mask into a cpl_mask structure and use it to set the
       internal mask for each of the input images */

    inbpm = vircam_mask_get_data(mask);
    nx = vircam_mask_get_size_x(mask);
    ny = vircam_mask_get_size_y(mask);
    npts = nx*ny;
    cplmask = cpl_mask_wrap((cpl_size)nx,(cpl_size)ny,(cpl_binary *)inbpm);
    for (i = 0; i < nfiles; i++)
    	cpl_image_reject_from_mask(vircam_fits_get_image(inlist[i]),cplmask);
    cpl_mask_unwrap(cplmask);

    /* For each file, do a quickie background analysis and reject high
       pixels */

    /* test = cpl_malloc(nx*ny*sizeof(float)); */
    /* for (i = 0; i < nfiles; i++) { */
    /* 	vircam_backmap(inlist[i],mask,64,&outim,&skymed); */
    /* 	odata = cpl_image_get_data_float(outim); */
    /* 	idata = cpl_image_get_data_float(vircam_fits_get_image(inlist[i])); */
    /* 	for (j = 0; j < npts; j++)  */
    /* 	    test[j] = idata[j] - odata[j]; */
    /* 	vircam_qmedsig(test,inbpm,(long)npts,3.0,2,-65535.0,65535.0,&skymed, */
    /* 		       &skysig); */
    /* 	immask = cpl_mask_get_data(cpl_image_get_bpm(vircam_fits_get_image(inlist[i]))); */
    /* 	for (j = 0; j < npts; j++)  */
    /* 	    if (test[j] > (skymed + 1.5*skysig))  */
    /* 		immask[j] = 1; */
    /* 	cpl_image_delete(outim); */
    /* } */
    /* freespace(test); */

    /* How many jitter steps are there? */

    (void)vircam_pfits_get_njsteps(vircam_fits_get_phu(inlist[0]),&njit);

    /* Allocate some workspce */

    xskys = cpl_malloc(njit*sizeof(xsky_struct));
    for (i = 0; i < njit; i++) {
	xskys[i].n = 0;
        xskys[i].in = cpl_malloc(8*sizeof(vir_fits *));
	xskys[i].nalloc = 8;
	xskys[i].xsky = NULL;
	xskys[i].status = VIR_OK;
    }
    xsky_fits = cpl_malloc(njit*sizeof(vir_fits *));

    /* Now loop through all the files and put them into the relevant 
       structure depending upon where they are in the jitter pattern.
       NB: We don't have to check on the status of the individual
       frames because only good frames should have been passed in by
       the calling routine. */

    for (i = 0; i < nfiles; i++) {
        (void)vircam_pfits_get_jitteri(vircam_fits_get_phu(inlist[i]),&ji);
	ji--;
	if (xskys[ji].n == xskys[ji].nalloc) {
	    xskys[ji].in = cpl_realloc(xskys[ji].in,
				       (xskys[ji].nalloc+8)*sizeof(vir_fits *));
	    xskys[ji].nalloc += 8;
        }
        xskys[ji].in[xskys[ji].n] = inlist[i];
	xskys[ji].n += 1;
    }
    
    /* Now loop through each of these groups and form a sky file with
       rejection */

    ns = 0;
    for (i = 0; i < njit; i++) {
	*status = VIR_OK;
	if (xskys[i].n == 0) {
	    xskys[i].xsky = NULL;
	    continue;
	} else if (xskys[i].n == 1) {
	    xskys[i].xsky = NULL;
	    xsky_fits[ns++] = xskys[i].in[0];
	    continue;
	}
	outim = NULL;
	(void)vircam_imcombine(xskys[i].in,xskys[i].n,1,1,0,1.0,&outim,&rejmask,
			       &rejplus,&drs,status);
	freespace(rejmask);
	freespace(rejplus);
	freepropertylist(drs);
	if (*status == VIR_OK) {
	    xskys[i].xsky = vircam_fits_wrap(outim,xskys[i].in[0],NULL,NULL);
	    xsky_fits[ns++] = xskys[i].xsky;
	} else {
	    freeimage(outim);
	}
    }

    /* Now combine all the intermediate skies to form the final sky */

    if (ns != 0) {
	outim = NULL;
        (void)vircam_imcombine(xsky_fits,ns,1,1,0,1.0,&outim,&rejmask,&rejplus,
			       &drs,status);
	freespace(rejmask);
	freespace(rejplus);
	freepropertylist(drs);
	if (*status == VIR_OK) {
	    *skyout = vircam_fits_wrap(outim,inlist[0],NULL,NULL);
	    (void)vircam_inpaint(*skyout,64,status);
	} else {
	    freeimage(outim);
	    *skyout = NULL;
	}
    } else {
	*skyout = NULL;
    }

    /* Tidy and exit */

    freespace(xsky_fits);
    for (i = 0; i < njit; i++) {
	freespace(xskys[i].in);
	freefits(xskys[i].xsky);
    }
    freespace(xskys);
    return(*status);
}


/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sky_mask_grow
    \par Purpose:
        Grow the bad pixels in a frame out to a given radius
    \par Description:
        The mask associated with an input image is modified to grow the
	bad pixels out to a given radius.
    \par Language:
        C
    \param dith
        The input image
    \param rad
        The grow radius
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sky_mask_grow(cpl_image *dith, float rad) {
    cpl_binary *inmap,*outmap;
    int nx,ny,ir,i,j,indx,ixmin,ixmax,iymin,iymax,ii,jj,indx2;
    float dx,dy,radius;

    /* Get the input map */

    inmap = cpl_mask_get_data(cpl_image_get_bpm(dith));
    nx = (int)cpl_image_get_size_x(dith);
    ny = (int)cpl_image_get_size_y(dith);

    /* Get an output map of the same size. Copy the input map to it */

    outmap = cpl_malloc(nx*ny*sizeof(*outmap));
    memmove(outmap,inmap,nx*ny*sizeof(*inmap));

    /* What is the minimum cell size that we need to consider given the 
       input radius */

    ir = vircam_nint(rad);

    /* Right, loop through the input image. If the current input pixel is
       not flagged, then move on to the next one. If it is, then look at
       a cell around it. Find all the pixels in that cell that are within
       the grow radius and flag them */

    for (j = 0; j < ny; j++) {
        for (i = 0; i < nx; i++) {
	    indx = j*nx + i;
	    if (! inmap[indx])
		continue;
	    ixmin = max(0,i-ir);
	    ixmax = min(nx-1,i+ir);
	    iymin = max(0,j-ir);
	    iymax = min(ny-1,j+ir);
	    for (jj = iymin; jj <= iymax; jj++) {
		dy = (float)(jj - j);
		for (ii = ixmin; ii <= ixmax; ii++) {
		    dx = (float)(ii - i);
		    radius = (float)sqrt(pow(dx,2.0) + pow(dy,2.0));
		    if (radius <= rad) { 
 		        indx2 = jj*nx + ii;
			outmap[indx2] = 1;
		    }
		}
	    }
	}
    }

    /* Now move the result back into the input mask. Free up workspace
       and get out of here */
   
    memmove(inmap,outmap,nx*ny*sizeof(*inmap));
    cpl_free(outmap);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sky_med
    \par Purpose:
        Work out the median of a sky frame
    \par Description:
        Work out the median of a sky frame taking the associated bad
	pixel mask into account
    \par Language:
        C
    \param sky
        The input sky frame
    \returns
        The background median
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static float vircam_sky_med(vir_fits *sky) {
    int npts;
    float *data,med;
    unsigned char *bpm;
    cpl_image *skyim;

    /* Get the size of the data array */

    skyim = vircam_fits_get_image(sky);
    npts = (int)cpl_image_get_size_x(skyim)*(int)cpl_image_get_size_y(skyim);

    /* Get the data */

    data = cpl_image_get_data_float(skyim);
    bpm = (unsigned char *)cpl_mask_get_data(cpl_image_get_bpm(skyim));

    /* Get the median */

    med = vircam_med(data,bpm,(long)npts);
    return(med);
}


/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sky_joffs
    \par Purpose:
        Work out the jitter offsets for a list of files
    \par Description:
        Work out the jitter offsets for a list of files using the WCS
	information in the headers
    \par Language:
        C
    \param inlist
        The input frame list
    \param nfiles
        The number of files in the input list
    \param xoffs
        The X coordinate offset relative to the first good image in the list
    \param yoffs
        The Y coordinate offset relative to the first good image in the list
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sky_joffs(vir_fits **inlist, int nfiles, float **xoffs,
			     float **yoffs) {
    float xoff,yoff;
    cpl_wcs *wcsref,*wcs;
    int refset,i,status;
    const double maxoffset = 2048;
    vir_fits *ff;
    const char *fctid = "vircam_sky_joffs";

    /* Get a bit of workspace to hold the offsets */

    *xoffs = cpl_malloc(nfiles*sizeof(float));
    *yoffs = cpl_malloc(nfiles*sizeof(float));

    /* Work out the jitter offsets from the WCS. First loop for each image */

    refset = 0;
    wcsref = NULL;
    for (i = 0; i < nfiles; i++) {
	ff = inlist[i];
	wcs = cpl_wcs_new_from_propertylist(vircam_fits_get_ehu(ff));

	/* If we can't get a WCS for this image, then signal that with
	   a warning */

	if (wcs == NULL) {
	    cpl_msg_warning(fctid,"Unable to get WCS for %s",
			    vircam_fits_get_filename(ff));
	    (*xoffs)[i] = 0.0;
	    (*yoffs)[i] = 0.0;
	    vircam_fits_set_error(ff,VIR_WARN);
	    continue;
	}

        /* Define the reference wcs pointer and set the first offset to zero */

	if (! refset) {
            (*xoffs)[0] = 0.0;
            (*yoffs)[0] = 0.0;
	    refset = 1;
	    wcsref = wcs;
	    continue;
	}

	/* Work out the x,y offset */

	status = VIR_OK;
	(void)vircam_diffxywcs(wcs,wcsref,&xoff,&yoff,&status);

	/* Did it work? If not the set a warning status for this file */

	if (status != VIR_OK) {
	    (*xoffs)[i] = 0.0;
	    (*yoffs)[i] = 0.0;
	    cpl_msg_warning(fctid,"Unable to WCS difference for %s",
			    vircam_fits_get_filename(ff));
        } else if (fabs((double)xoff) > maxoffset || 
		   fabs((double)yoff) > maxoffset) {
	    vircam_fits_set_error(ff,VIR_FATAL);
	    cpl_msg_error(fctid,"WCS offsets for %s are >%g: %g %g -- ignoring",
			  vircam_fits_get_filename(ff),maxoffset,xoff,yoff);
	} else {
	    (*xoffs)[i] = xoff;
	    (*yoffs)[i] = yoff;
	}
	cpl_wcs_delete(wcs);
    }
    if (wcsref != NULL) 
	cpl_wcs_delete(wcsref);

    /* Write the results to the headers */

    for (i = 0; i < nfiles; i++) {
	ff = inlist[i];
	cpl_propertylist_update_double(vircam_fits_get_ehu(ff),
				       "ESO DRS XOFFDITHER",
				       (double)(*xoffs)[i]);
	cpl_propertylist_update_double(vircam_fits_get_ehu(ff),
				       "ESO DRS YOFFDITHER",
				       (double)(*yoffs)[i]);
    }
}

static void vircam_sky_xytoxy(cpl_wcs *inwcs, cpl_wcs *outwcs, 
			      cpl_matrix *inxy, cpl_matrix **outxy) {
    cpl_matrix *radec;
    cpl_array *wstatus;

    /* Convert the input xy data to RA/Dec */

    cpl_wcs_convert(inwcs,inxy,&radec,&wstatus,CPL_WCS_PHYS2WORLD);
    cpl_array_delete(wstatus);

    /* Now convert the RA/Dec to the xy coordinates of the second image */

    cpl_wcs_convert(outwcs,radec,outxy,&wstatus,CPL_WCS_WORLD2PHYS);
    cpl_array_delete(wstatus);
    cpl_matrix_delete(radec);
}
    
/**@}*/

/*

$Log: not supported by cvs2svn $
Revision 1.9  2012/01/27 12:25:10  jim
Fixed some casts

Revision 1.8  2012/01/15 17:40:09  jim
Minor modifications to take into accout the changes in cpl API for v6

Revision 1.7  2010/06/30 12:42:00  jim
A few fixes to stop compiler compaints

Revision 1.6  2009/02/23 10:46:26  jim
Modified tilesky to try and get rid of some of the low level glow around
bright and extended objects

Revision 1.5  2009/02/20 11:01:13  jim
Added vircam_tilesky. Commented out vircam_skycombine as it's not being
used at present and it's generating a compiler error. May delete it
completely later

Revision 1.4  2008/11/25 06:23:09  jim
Added some routine prologues

Revision 1.3  2008/11/21 10:12:36  jim
Fixed typo

Revision 1.2  2008/10/24 10:57:19  jim
Fixed bug in pawsky_mask so that the bad pixels in the confidence map are
folded into the object mask

Revision 1.1  2008/10/13 08:13:21  jim
New entry

*/
